# ICLR submission

This repository contains code for the ICLR submission #13884 titled "Language Models are Advanced Anonymizers"

## Setup

Install mamba via:

```
curl -L -O "https://github.com/conda-forge/miniforge/releases/latest/download/Mambaforge-$(uname)-$(uname -m).sh"
bash Mambaforge-$(uname)-$(uname -m).sh
```

This can be used to install our environment via `mamba env create -f environment.yaml`

Afterward, you should edit the file `credentials.py` to contain your OpenAI and Azure keys, respectively.

Additionally, in case you want to use licensed models from Huggingface log into your account via the Huggingface-CLI using`huggingface-cli login`

## Important notes before running

Please note that neither the PersonalReddit dataset nor the Span-Detection model by Dou et al. are open-access. Therefore, both have been removed from this repository. The repository does, however, contain synthetic examples from "Beyond Memorization" (Staab et al., 2023). Corresponding configuration files can be found under `configs/anonymization/synthetic`.

## Running

Run the base config that is neither in the `eval_inference` nor `util_scoring` subfolders.

In case your locally applied judge is already GPT-4 you can skip the eval_inference config. Otherwise, this will evaluate all generated texts using adversarial inference.

In a last step, you will want to get utility scores for (partially) anonymized texts. For this, you can use the respective configs in in `../eval_inference/<config>`.

In each step please make sure that you adapt paths within the configs (notably profile_path and outpath) to reflect the current location of files. (Side note: You will find that as a cost-saving measure, we shared the inferences on fully non-anonymized text).

## Evaluating

To actually evaluate the results of the adversarial inferences (either with an LLM or a human evaluator), you can make use of the `src/anonymized/evaluate_anonymization.py` script.

In particular, you want to run it as follows:

`python src/anonymized/evaluate_anonymization.py --in_path <your_eval_inference_with_utility>.jsonl --decider "model_human" --out_path <out_directory>` --score

to create a canonical `eval_out.jsonl`

running the same command again without the `--score` will translate this into the csv format used in our plotting script.

## Plotting

All our plots have been created with a single call to the `all_plots.sh` script. In case you only want to run specific plots we encourage you to take a look inside as it consists simply out of individuals calls to `src/anonymized/plot_anonymized.py`.
